from pathlib import Path
import prompt_toolkit
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.validation import Validator

from tala.cli.tdm.passivity_timer import PassivityTimer, PASSIVE
from tala.utils.tdm_client import TDMClient
from tala.utils.observable import Observer


class TDMCLI(object):
    def __init__(self, url):
        self._client = TDMClient(url)
        self._tdm_session = None
        self._passivity_timer = PassivityTimer()
        self._passivity_timer.add_observer(PassivityTimerObserver(self))
        history = self._create_file_history()
        activity_detector = self._create_activity_detector_as_empty_validator()
        self._session = PromptSession(history=history, mouse_support=True, validator=activity_detector)

    def _create_activity_detector_as_empty_validator(self):
        def on_active(text):
            if text:
                self._passivity_timer.stop()
            return True

        return Validator.from_callable(on_active)

    def stop(self):
        self._passivity_timer.stop()

    def run(self):
        for system_turn in self.system_turns():
            print system_turn

    def system_turns(self):
        greeting = self._request(self._client.start_session)
        yield greeting
        while True:
            message = self._session.prompt(u"U> ").rstrip()
            self._passivity_timer.stop()
            yield self._request(self._client.request_text_input, self._tdm_session, message)

    def _request(self, callable, *args, **kwargs):
        response = callable(*args, **kwargs)
        if not self._tdm_session:
            self._tdm_session = response["session"]["session_id"]
        expected_passivity = response["output"]["expected_passivity"]
        if expected_passivity is not None:
            self._passivity_timer.start(expected_passivity)
        system_utterance = response["output"]["utterance"]
        return u"S> {}".format(system_utterance)

    def request_passivity(self):
        system_utterance = self._request(self._client.request_passivity, self._tdm_session)

        def print_passivity_turn():
            print u"U>"
            print system_utterance

        prompt_toolkit.application.run_in_terminal(print_passivity_turn)

    def _create_file_history(self):
        history_file = Path(".tala") / "history"
        if not history_file.parent.exists():
            history_file.parent.mkdir()
        if not history_file.exists():
            history_file.touch()
        return FileHistory(str(history_file))


class PassivityTimerObserver(Observer):
    def __init__(self, tdm_cli):
        self._tdm_cli = tdm_cli

    def update(self, value):
        if value == PASSIVE:
            self._tdm_cli.request_passivity()



import json

from mock import patch, Mock, call
import pytest

from tala.model.input_hypothesis import InputHypothesis
from tala.utils import tdm_client
from tala.utils.tdm_client import TDMClient, PROTOCOL_VERSION, TDMRuntimeException


class TestTDMClient(object):
    def setup(self):
        self._mocked_requests = None
        self._tdm_client = None
        self._session = None

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_start_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._when_calling_start_session()
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": {version}, "request": {{"type": "start_session"}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _given_mocked_requests_module(self, mock_requests):
        self._mocked_requests = mock_requests

    def _given_tdm_client_created_for(self, url):
        self._tdm_client = TDMClient(url)

    def _when_calling_start_session(self):
        self._tdm_client.start_session()

    def _then_request_was_made_with(self, url, headers, data):
        self._mocked_requests.post.assert_has_calls([call(url, data=data, headers=headers)])

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_say_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_text_input("mock-utterance")
        self._then_request_was_made_with(
            "mock-url",
            data='{{"session": {{"session_id": "mock-session-id"}}, '
            '"input": {{"utterance": "mock-utterance", "modality": "text"}}, '
            '"version": {version}, "request": {{"type": "input"}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _given_session_id(self, session_id):
        self._session = session_id

    def _given_requests_respond_with(self, session_id):
        response = Mock()
        json_response = '''{{
  "version": {version},
  "session": {{
    "session_id": "{session_id}"
  }},
  "output": {{
    "utterance": "mock-system-utterance",
    "expected_passivity": 5.0,
    "actions": []
  }},
  "nlu_result": {{
    "selected_utterance": "call John",
    "confidence": 0.88
  }},
  "context": {{
    "active_ddd": "my_ddd",
    "facts": {{}},
    "language": "eng"
  }}
}}'''.format(
            version=PROTOCOL_VERSION, session_id=session_id
        )
        response.json.return_value = json.loads(json_response)
        self._mocked_requests.post.return_value = response

    def _when_requesting_text_input(self, utterance):
        self._tdm_client.request_text_input(self._session, utterance)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_requesting_speech_input_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_speech_input([InputHypothesis("mock-utterance", "mock-confidence")])
        self._then_request_was_made_with(
            "mock-url",
            data='{{"session": {{"session_id": "mock-session-id"}}, '
            '"input": {{"hypotheses": [{{"confidence": "mock-confidence", "utterance": "mock-utterance"}}], '
            '"modality": "speech"}}, '
            '"version": {version}, "request": {{"type": "input"}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _when_requesting_speech_input(self, hypotheses):
        self._tdm_client.request_speech_input(self._session, hypotheses)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_passivity_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_calling_passivity()
        self._then_request_was_made_with(
            "mock-url",
            data='{{"session": {{"session_id": "mock-session-id"}}, '
            '"version": {version}, "request": {{"type": "passivity"}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _when_calling_passivity(self):
        self._tdm_client.request_passivity(self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_start_session_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_requests_error_with("mock-error")
        self._when_starting_session_then_exception_was_raised(TDMRuntimeException, "mock-error")

    def _when_starting_session_then_exception_was_raised(self, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.start_session()

    def _given_requests_error_with(self, description):
        response = Mock()
        json_response = '''{{
  "version": {version},
  "session": {{
    "session_id": "mock-session"
  }},
  "error": {{
    "description": "{description}"
  }}
}}'''.format(
            version=PROTOCOL_VERSION, description=description
        )
        response.json.return_value = json.loads(json_response)
        self._mocked_requests.post.return_value = response

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_say_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_text_input_then_exception_was_raised("mock-utterance", TDMRuntimeException, "mock-error")

    def _when_requesting_text_input_then_exception_was_raised(self, utterance, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_text_input(self._session, utterance)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_request_passivity_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_passivity_then_exception_was_raised(TDMRuntimeException, "mock-error")

    def _when_requesting_passivity_then_exception_was_raised(self, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_passivity(self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_request_speech_input_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_speech_input_then_exception_was_raised([
            InputHypothesis("mock-utterance", "mock-confidence")
        ], TDMRuntimeException, "mock-error")

    def _when_requesting_speech_input_then_exception_was_raised(self, hypotheses, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_speech_input(self._session, hypotheses)



import warnings

from tala.backend.dependencies.abstract_backend_dependencies import AbstractBackendDependencies
from tala.ddd.loading.ddd_set_loader import DDDSetLoader
from tala.model.sort import DATETIME, INTEGER


class BackendDependenciesForGenerating(AbstractBackendDependencies):
    def __init__(self, backend_args):
        super(BackendDependenciesForGenerating, self).__init__(backend_args)
        self._validate_rasa_compatibility()

    def _validate_rasa_compatibility(self):
        for ddd in self.ddds:
            if ddd.ontology.predicates_contain_sort(DATETIME):
                warnings.warn(
                    "DDD '{0}' contains predicates of sort '{1}'. "
                    "Note that the builtin NLU can not recognize them. "
                    "For best coverage, use a properly configured Rasa NLU.".format(ddd.name, DATETIME)
                )
            if ddd.ontology.predicates_contain_sort(INTEGER):
                warnings.warn(
                    "DDD '{0}' contains predicates of sort '{1}'. "
                    "Note that the builtin NLU has limitations with this sort. "
                    "For best coverage, use a properly configured Rasa NLU.".format(ddd.name, INTEGER)
                )

    def load_ddds(self, ddd_names):
        ddd_set_loader = self._create_ddd_set_loader()
        ddds = ddd_set_loader.ddds_as_list(ddd_names, languages=self.supported_languages)
        return ddds

    def _create_ddd_set_loader(self):
        return DDDSetLoader(self.overridden_ddd_config_paths)



import unittest
import warnings

from mock import patch, call

from tala.backend import dependencies
from tala.backend.dependencies.for_generating import BackendDependenciesForGenerating
from tala.model.sort import DATETIME, INTEGER
from tala.testing.backend_dependencies_test_base import BackendDependenciesTestBase


class BackendDependenciesForGeneratingTestCase(BackendDependenciesTestBase, unittest.TestCase):
    def setUp(self):
        super(BackendDependenciesForGeneratingTestCase, self).setUp()
        self._config = None
        self._backend_dependencies = None
        self._rasa_config = None
        self._mock_rasa_interpreter = None
        self._mock_rasa_loader = None

    @patch("%s.abstract_backend_dependencies.BackendConfig" % dependencies.__name__, autospec=True)
    @patch("%s.for_generating.DDDSetLoader" % dependencies.__name__, autospec=True)
    def test_datetime_predicates_in_ontology_issues_warning(self, DddSetLoader, BackendConfig):
        self.given_mock_backend_config(BackendConfig)
        self.given_mock_ddd_set_loader(DddSetLoader)
        self.given_ddds(["mock_ddd"])
        self.given_active_ddd_in_config("mock_ddd")
        self.given_mocked_ontology_in("mock_ddd")
        self.given_mocked_ontology_has_predicates_of_sort("mock_ddd", [DATETIME])
        self.when_create_backend_dependencies_then_warning_is_issued_matching([
            "DDD 'mock_ddd' contains predicates of sort 'datetime'. "
            "Note that the builtin NLU can not recognize them. "
            "For best coverage, use a properly configured Rasa NLU."
        ])

    def when_create_backend_dependencies_then_warning_is_issued_matching(self, expected_warnings):
        with patch("{}.for_generating.warnings".format(dependencies.__name__), spec=warnings) as mock_warnings:
            self._create_backend_dependencies()
            mock_warnings.warn.assert_has_calls([call(warning) for warning in expected_warnings])

    @patch("%s.abstract_backend_dependencies.BackendConfig" % dependencies.__name__, autospec=True)
    @patch("%s.for_generating.DDDSetLoader" % dependencies.__name__, autospec=True)
    def test_integer_predicates_in_ontology_issues_warning(self, DddSetLoader, BackendConfig):
        self.given_mock_backend_config(BackendConfig)
        self.given_mock_ddd_set_loader(DddSetLoader)
        self.given_ddds(["mock_ddd"])
        self.given_active_ddd_in_config("mock_ddd")
        self.given_mocked_ontology_in("mock_ddd")
        self.given_mocked_ontology_has_predicates_of_sort("mock_ddd", [INTEGER])
        self.when_create_backend_dependencies_then_warning_is_issued_matching([
            "DDD 'mock_ddd' contains predicates of sort 'integer'. "
            "Note that the builtin NLU has limitations with this sort. "
            "For best coverage, use a properly configured Rasa NLU."
        ])

    def given_mocked_ontology_has_no_predicates_of_any_sorts(self, ddd):
        self._get_ddd(ddd).ontology.predicates_contain_sort.return_value = False

    def when_create_backend_dependencies_then_no_warnings_are_issued(self):
        with patch("{}.for_generating.warnings".format(dependencies.__name__), spec=warnings) as mock_warnings:
            self._create_backend_dependencies()
            mock_warnings.warn.assert_not_called()

    @patch("%s.abstract_backend_dependencies.BackendConfig" % dependencies.__name__, autospec=True)
    @patch("%s.for_generating.DDDSetLoader" % dependencies.__name__, autospec=True)
    def test_integer_and_datetime_predicates_in_ontology_issues_warning(self, DddSetLoader, BackendConfig):
        self.given_mock_backend_config(BackendConfig)
        self.given_mock_ddd_set_loader(DddSetLoader)
        self.given_ddds(["mock_ddd"])
        self.given_active_ddd_in_config("mock_ddd")
        self.given_mocked_ontology_in("mock_ddd")
        self.given_mocked_ontology_has_predicates_of_sort("mock_ddd", [INTEGER, DATETIME])
        self.when_create_backend_dependencies_then_warning_is_issued_matching([
            "DDD 'mock_ddd' contains predicates of sort 'datetime'. "
            "Note that the builtin NLU can not recognize them. "
            "For best coverage, use a properly configured Rasa NLU.",
            "DDD 'mock_ddd' contains predicates of sort 'integer'. "
            "Note that the builtin NLU has limitations with this sort. "
            "For best coverage, use a properly configured Rasa NLU.",
        ])

    @patch("%s.abstract_backend_dependencies.BackendConfig" % dependencies.__name__, autospec=True)
    @patch("%s.for_generating.DDDSetLoader" % dependencies.__name__, autospec=True)
    def test_no_predicates_of_rasa_dependent_sorts_in_ontology_issues_no_warnings(self, DddSetLoader, BackendConfig):
        self.given_mock_backend_config(BackendConfig)
        self.given_mock_ddd_set_loader(DddSetLoader)
        self.given_ddds(["mock_ddd"])
        self.given_active_ddd_in_config("mock_ddd")
        self.given_mocked_ontology_in("mock_ddd")
        self.given_mocked_ontology_has_no_predicates_of_any_sorts("mock_ddd")
        self.when_create_backend_dependencies_then_no_warnings_are_issued()

    def _create_backend_dependencies(self):
        self._backend_dependencies = BackendDependenciesForGenerating(self._mock_args)



import json

import requests

from tala.utils.observable import Observable

PROTOCOL_VERSION = 2.0


class InvalidResponseError(Exception):
    pass


class TDMRuntimeException(Exception):
    pass


class TDMClient(Observable):
    def __init__(self, url):
        super(TDMClient, self).__init__()
        self._url = url

    def request_text_input(self, session, utterance):
        request = self._create_text_input_request(session, utterance)
        response = self._make_request(request)
        return response

    def request_speech_input(self, session, hypotheses):
        request = self._create_speech_input_request(session, hypotheses)
        response = self._make_request(request)
        return response

    def request_passivity(self, session):
        request = {"version": PROTOCOL_VERSION, "session": {"session_id": session}, "request": {"type": "passivity"}}
        response = self._make_request(request)
        return response

    def _create_text_input_request(self, session, utterance):
        return {
            "version": PROTOCOL_VERSION,
            "session": {
                "session_id": session
            },
            "request": {
                "type": "input"
            },
            "input": {
                "modality": "text",
                "utterance": utterance,
            }
        }

    def _create_speech_input_request(self, session, hypotheses):
        return {
            "version": PROTOCOL_VERSION,
            "session": {
                "session_id": session
            },
            "request": {
                "type": "input"
            },
            "input": {
                "modality":
                "speech",
                "hypotheses": [{
                    "utterance": hypothesis.utterance,
                    "confidence": hypothesis.confidence,
                } for hypothesis in hypotheses],
            }
        }

    def start_session(self):
        request = {"version": PROTOCOL_VERSION, "request": {"type": "start_session"}}
        response = self._make_request(request)
        return response

    def _make_request(self, request_body):
        data_as_json = json.dumps(request_body)
        headers = {'Content-type': 'application/json'}
        json_encoded_response = requests.post(self._url, data=data_as_json, headers=headers)
        try:
            response = json_encoded_response.json()
        except ValueError:
            raise InvalidResponseError("Expected a valid JSON response but got %s." % json_encoded_response)

        if "error" in response:
            description = response["error"]["description"]
            raise TDMRuntimeException(description)

        return response



#!/usr/bin/env python2.7

import argparse
import contextlib
import os
import re
import sys
import warnings

from jinja2 import Template
from pathlib import Path

from tala.backend.dependencies.for_generating import BackendDependenciesForGenerating
from tala.cli.argument_parser import add_common_backend_arguments
from tala.cli.tdm.tdm_cli import TDMCLI
from tala.config import BackendConfig, DddConfig, DeploymentsConfig, BackendConfigNotFoundException, \
    DddConfigNotFoundException, DeploymentsConfigNotFoundException
from tala.ddd.building.ddd_builder_for_generating import DDDBuilderForGenerating
from tala import installed_version
from tala.cli import console_formatting
from tala.nl import languages
from tala.nl.rasa.generating.generator import RasaGenerator
from tala.ddd.maker import utils as ddd_maker_utils
from tala.ddd.maker.ddd_maker import DddMaker
from tala.utils.chdir import chdir


class ConfigAlreadyExistsException(Exception):
    pass


class ConfigNotFoundException(Exception):
    pass


class InvalidArgumentException(Exception):
    pass


class UnexpectedDDDException(Exception):
    pass


class UnexpectedEnvironmentException(Exception):
    pass


def create_ddd(args):
    def directory_to_class_name(directory_name):
        name_with_capitalized_words = directory_name.title()
        class_name = re.sub("[_ ]", "", name_with_capitalized_words)
        return class_name

    directory_name = args.name
    class_name = directory_to_class_name(directory_name)
    DddMaker(class_name, directory_name, args.target_dir).make()


def create_backend_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            "Expected to be able to create backend config file '%s' but it already exists." % args.filename
        )
    BackendConfig().write_default_config(args.filename, args.ddd)


def create_ddd_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            "Expected to be able to create DDD config file '%s' but it already exists." % args.filename
        )
    DddConfig().write_default_config(args.filename)


def create_deployments_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            "Expected to be able to create deployments config file '%s' but it already exists." % args.filename
        )
    DeploymentsConfig().write_default_config(args.filename)


def verify(args):
    backend_dependencies = BackendDependenciesForGenerating(args)

    ddds_builder = DDDBuilderForGenerating(
        backend_dependencies,
        verbose=args.verbose,
        ignore_warnings=args.ignore_warnings,
        language_codes=args.language_codes
    )
    ddds_builder.verify()


def generate_rasa(args):
    backend_dependencies = BackendDependenciesForGenerating(args)
    ddd_path = Path(args.ddd)
    if not ddd_path.exists():
        raise UnexpectedDDDException("Expected DDD '{}' to exist but it didn't".format(args.ddd))

    with chdir(ddd_path / "grammar"):
        generator = RasaGenerator(backend_dependencies.ddds[0], args.language)
        data = generator.generate()
        language = languages.RASA_LANGUAGE[args.language]

        rasa_config_template = Template(
            "language: \"{{ language }}\"\n"
            "\n"
            "pipeline: \"spacy_sklearn\"\n"
            "\n"
            "data: |\n"
            "{% for line in data.splitlines() %}"
            "  {{ line }}\n"
            "{% endfor %}"
        )

        rasa_config_template.stream(language=language, data=data).dump(sys.stdout, encoding="utf-8")


def _check_ddds_for_word_lists(ddds):
    for ddd in ddds:
        ddd_maker_utils.potentially_create_word_list_boilerplate(ddd.name)


@contextlib.contextmanager
def _config_exception_handling():
    def generate_message(name, command_name, config):
        return "Expected {name} config '{config}' to exist but it was not found. To create it, " \
               "run 'tala {command} --filename {config}'.".format(name=name, command=command_name, config=config)

    try:
        yield
    except BackendConfigNotFoundException as exception:
        message = generate_message("backend", "create-backend-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DddConfigNotFoundException as exception:
        message = generate_message("DDD", "create-ddd-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DeploymentsConfigNotFoundException as exception:
        message = generate_message("deployments", "create-deployments-config", exception.config_path)
        raise ConfigNotFoundException(message)


def version(args):
    print installed_version.version


def interact(args):
    def get_url():
        config = DeploymentsConfig(args.deployments_config).read()
        if args.environment not in config:
            raise UnexpectedEnvironmentException(
                "Expected one of the known environments {} but got '{}'".format(config.keys(), args.environment)
            )
        return config[args.environment]

    url = get_url()
    tdm_cli = TDMCLI(url)

    try:
        tdm_cli.run()
    except (KeyboardInterrupt, EOFError):
        tdm_cli.stop()


def add_verify_subparser(subparsers):
    parser = subparsers.add_parser(
        "verify", help="verify the format of all DDDs supported by the backend, across all supported languages"
    )
    parser.set_defaults(func=verify)
    add_common_backend_arguments(parser)
    parser.add_argument(
        "--languages",
        dest="language_codes",
        choices=languages.SUPPORTED_LANGUAGES,
        nargs="*",
        help="override the backend config and verify the DDDs for these languages"
    )
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose")
    parser.add_argument(
        "--ignore-grammar-warnings",
        dest="ignore_warnings",
        action="store_true",
        help="ignore warnings when compiling the grammar"
    )


def add_generate_rasa_subparser(subparsers):
    parser = subparsers.add_parser("generate-rasa", help="generate training data for Rasa NLU 0.14")
    parser.set_defaults(func=generate_rasa)
    add_common_backend_arguments(parser)
    parser.add_argument("ddd", help="generate for this DDD")
    parser.add_argument("language", choices=languages.SUPPORTED_RASA_LANGUAGES, help="use the grammar of this language")


def add_create_ddd_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd", help="create a new DDD")
    parser.set_defaults(func=create_ddd)
    parser.add_argument("name", help="Name of the DDD, e.g. my_ddd")
    parser.add_argument(
        "--target-dir", default=".", metavar="DIR", help="target directory, will be created if it doesn't exist"
    )


def add_create_backend_config_subparser(subparsers):
    parser = subparsers.add_parser("create-backend-config", help="create a backend config")
    parser.set_defaults(func=create_backend_config)
    parser.add_argument(
        "--filename",
        default=BackendConfig.default_name(),
        metavar="NAME",
        help="filename of the backend config, e.g. %s" % BackendConfig.default_name()
    )
    parser.add_argument("--ddd", help="name of the active DDD, e.g. my_ddd")


def add_create_ddd_config_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd-config", help="create a DDD config")
    parser.set_defaults(func=create_ddd_config)
    parser.add_argument(
        "--filename",
        default=DddConfig.default_name(),
        metavar="NAME",
        help="filename of the DDD config, e.g. %s" % DddConfig.default_name()
    )


def add_create_deployments_config_subparser(subparsers):
    parser = subparsers.add_parser("create-deployments-config", help="create a deployments config")
    parser.set_defaults(func=create_deployments_config)
    parser.add_argument(
        "--filename",
        default=DeploymentsConfig.default_name(),
        metavar="NAME",
        help="filename of the deployments config, e.g. %s" % DeploymentsConfig.default_name()
    )


def add_version_subparser(subparsers):
    parser = subparsers.add_parser("version", help="print the Tala version")
    parser.set_defaults(func=version)


def add_interact_subparser(subparsers):
    parser = subparsers.add_parser("interact", help="start an interactive chat with a deployed DDD")
    parser.add_argument(
        "environment",
        help="specify the environment to use, e.g. 'dev'; this maps to the keys of the deployments config"
    )
    parser.add_argument(
        "--config",
        dest="deployments_config",
        default=None,
        help="override the default deployments config %r" % DeploymentsConfig.default_name()
    )
    parser.set_defaults(func=interact)


def format_warnings():
    def warning_on_one_line(message, category, _filename, _lineno, _file=None, _line=None):
        string = "%s: %s\n" % (category.__name__, message)
        return console_formatting.bold(string)

    warnings.formatwarning = warning_on_one_line


def show_deprecation_warnings():
    warnings.simplefilter("always", category=DeprecationWarning)


def main(args=None):
    format_warnings()
    show_deprecation_warnings()
    root_parser = argparse.ArgumentParser(description="Use the Tala SDK for the Talkamatic Dialogue Manager (TDM).")
    subparsers = root_parser.add_subparsers()
    add_verify_subparser(subparsers)
    add_generate_rasa_subparser(subparsers)
    add_create_ddd_subparser(subparsers)
    add_create_backend_config_subparser(subparsers)
    add_create_ddd_config_subparser(subparsers)
    add_create_deployments_config_subparser(subparsers)
    add_version_subparser(subparsers)
    add_interact_subparser(subparsers)

    parsed_args = root_parser.parse_args(args)
    with _config_exception_handling():
        parsed_args.func(parsed_args)


if __name__ == "__main__":
    main()



import os
import warnings

from jinja2 import Template

from tala.ddd.grammar.reader import GrammarReader
from tala.nl.rasa.constants import ACTION_INTENT, QUESTION_INTENT, ANSWER_INTENT, ANSWER_NEGATION_INTENT, \
    NEGATIVE_INTENT
from tala.nl.rasa.generating.examples import Examples


class RasaNotSupportedByGrammarFormatException(Exception):
    pass


class UnexpectedRequiredEntityException(Exception):
    pass


class UnexpectedPropositionalEntityEncounteredException(Exception):
    pass


BUILTIN_ENTITY_TEMPLATE = Template("{{ grammar_entry }}")
SORTAL_ENTITY_TEMPLATE = Template("[{{ grammar_entry }}](sort.{{ value }})")
PROPOSITIONAL_ENTITY_TEMPLATE = Template("[{{ grammar_entry }}](predicate.{{ value }})")


class RasaGenerator(object):
    def __init__(self, ddd, language_code):
        super(RasaGenerator, self).__init__()
        self._ddd = ddd
        self._language_code = language_code
        self._language_examples = Examples.from_language(language_code)

    def generate(self):
        def add(examples, new_examples):
            for key, values in new_examples.items():
                if key in examples:
                    examples[key].extend(values)
                else:
                    examples[key] = values

        if not GrammarReader.xml_grammar_exists_for_language(self._language_code):
            raise RasaNotSupportedByGrammarFormatException(
                "Expected an XML-based grammar at '%s', but it does not exist" %
                os.path.abspath(GrammarReader.path(self._language_code))
            )
        grammar = self._ddd.grammars[self._language_code]

        examples = {}
        add(examples, self._examples_of_requests(grammar))
        add(examples, self._examples_of_questions(grammar))
        add(examples, self._examples_of_sortal_answers_from_individuals(grammar))
        add(examples, self._examples_of_answers(grammar))
        add(examples, self._examples_of_sortal_answer_negations_from_individuals(grammar))

        data_template = Template(
            "{% for intent, examples in ddd_examples.items() %}"
            "{% if examples %}"
            "## intent:{{ ddd }}:{{ intent }}\n"
            "{% for example in examples %}"
            "- {{ example }}\n"
            "{% endfor %}"
            "\n"
            "{% endif %}"
            "{% endfor %}"
            ""
            "{% for intent, examples in general_examples.items() %}"
            "{% if examples %}"
            "## intent:{{ intent }}\n"
            "{% for example in examples %}"
            "- {{ example }}\n"
            "{% endfor %}"
            "\n"
            "{% endif %}"
            "{% endfor %}"
            ""
            "{% for synonym_object in synonym_objects %}"
            "## synonyms:{{ ddd }}:{{ synonym_object.value }}\n"
            "{% for synonym in synonym_object.synonyms %}"
            "- {{ synonym }}\n"
            "{% endfor %}"
            "\n"
            "{% endfor %}"
        )
        synonyms = self._entity_synonyms_from_custom_sorts(grammar)
        rasa_data = data_template.render(
            ddd_examples=examples,
            general_examples=self._examples_of_negative_intent(),
            synonym_objects=synonyms,
            ddd=self._ddd.name
        )

        return rasa_data

    def _entity_synonyms_from_custom_sorts(self, grammar):
        for sort in self._ddd.ontology.get_sorts().values():
            if sort.is_builtin():
                continue
            entities = list(self._all_individual_grammar_entries_of_custom_sort(grammar, sort))
            for individual_entities in entities:
                if len(individual_entities) <= 1:
                    continue
                yield self._create_synonym_object(individual_entities[0], individual_entities[1:])

    @staticmethod
    def _create_synonym_object(value, synonyms):
        return {
            "value": value,
            "synonyms": synonyms,
        }

    def _examples_of_negative_intent(self):
        return {NEGATIVE_INTENT: list(self._language_examples.negative)}

    def _examples_of_requests(self, grammar):
        examples = {}
        for action in self._ddd.ontology.get_ddd_specific_actions():
            intent = "%s::%s" % (ACTION_INTENT, action)
            examples[intent] = self._examples_of_request(grammar, action)
        return examples

    def _examples_of_request(self, grammar, action):
        requests = grammar.requests_of_action(action)
        for request in requests:
            for example in self._examples_of_intent(grammar, request):
                yield example

    def _examples_of_intent(self, grammar, intent):
        head = intent.text_chunks[0]
        texts = intent.text_chunks[1:]
        try:
            examples = self._examples_with_individuals(grammar, texts, intent.required_entities, [head])
        except UnexpectedPropositionalEntityEncounteredException:
            return
        for example in examples:
            yield example

    def _examples_with_individuals(self, grammar, text_chunks, required_entities, examples_so_far):
        if not text_chunks and not required_entities:
            return examples_so_far
        tail = text_chunks[0]
        required_entity = required_entities[0]
        all_new_examples = []
        for example in examples_so_far:
            if required_entity.is_sortal:
                new_examples = list(self._examples_from_sortal_individual(grammar, required_entity, example, tail))
                all_new_examples.extend(new_examples)
            elif required_entity.is_propositional:
                predicate = self._ddd.ontology.get_predicate(required_entity.name)
                if predicate.getSort().is_string_sort():
                    new_examples = list(
                        self._examples_from_propositional_individual(grammar, required_entity, example, tail)
                    )
                    all_new_examples.extend(new_examples)

                else:
                    message = (
                        "Expected only sortal slots but got a propositional slot for predicate '{}'. "
                        "Skipping this training data example.".format(predicate.get_name(), predicate.getSort())
                    )
                    warnings.warn(message, UserWarning)
                    raise UnexpectedPropositionalEntityEncounteredException(message)
            else:
                raise UnexpectedRequiredEntityException(
                    "Expected either a sortal or propositional required entity but got a %s" %
                    required_entity.__class__.__name__
                )
        return self._examples_with_individuals(grammar, text_chunks[1:], required_entities[1:], all_new_examples)

    def _examples_from_sortal_individual(self, grammar, required_sortal_entity, example_so_far, tail):
        sort = self._ddd.ontology.get_sort(required_sortal_entity.name)
        individuals = self._individual_grammar_entries_samples(grammar, sort)
        sort_should_be_generated_as_entity = sort.is_string_sort() or not sort.is_builtin()
        template = SORTAL_ENTITY_TEMPLATE if sort_should_be_generated_as_entity else BUILTIN_ENTITY_TEMPLATE
        return self._examples_from_individuals(
            template, sort.get_name(), individuals, example_so_far, tail
        )

    def _sample_individuals_of_builtin_sort(self, sort):
        examples = self._language_examples.get_builtin_sort_examples(sort)
        return [[entry] for entry in examples]

    def _individual_grammar_entries_samples(self, grammar, sort):
        if sort.is_builtin():
            return self._sample_individuals_of_builtin_sort(sort)
        return self._individual_grammar_entries_samples_of_custom_sort(grammar, sort)

    def _individual_grammar_entries_samples_of_custom_sort(self, grammar, sort):
        for individual_entries in self._all_individual_grammar_entries_of_custom_sort(grammar, sort):
            yield individual_entries

    def _all_individual_grammar_entries_of_custom_sort(self, grammar, sort):
        individuals = self._ddd.ontology.get_individuals_of_sort(sort.get_name())
        for individual in individuals:
            yield grammar.entries_of_individual(individual)

    def _examples_from_propositional_individual(self, grammar, required_propositional_entity, example_so_far, tail):
        predicate_name = required_propositional_entity.name
        predicate = self._ddd.ontology.get_predicate(predicate_name)
        sort = predicate.getSort()
        individuals = self._individual_grammar_entries_samples(grammar, sort)
        predicate_specific_samples = self._string_examples_of_predicate(grammar, predicate)
        individuals.extend([[predicate_specific_sample] for predicate_specific_sample in predicate_specific_samples])
        return self._examples_from_individuals(
            PROPOSITIONAL_ENTITY_TEMPLATE, predicate_name, individuals, example_so_far, tail
        )

    def _string_examples_of_predicate(self, grammar, predicate):
        return grammar.strings_of_predicate(predicate.get_name())

    def _examples_from_individuals(self, template, identifier, individuals_grammar_entries, example_so_far, tail):
        for grammar_entries in individuals_grammar_entries:
            for grammar_entry in grammar_entries:
                entity = template.render(grammar_entry=grammar_entry, value=identifier)
                example = self._extend_example(entity, example_so_far, tail)
                yield example

    @staticmethod
    def _extend_example(entity, example_so_far, tail=None):
        head = example_so_far
        tail = tail or ""
        return "".join([head, entity, tail])

    def _examples_of_questions(self, grammar):
        def examples(question):
            for example in self._examples_of_question(grammar, question):
                yield example

        result = {}
        for resolve_goal in self._ddd.domain.get_all_resolve_goals():
            question = resolve_goal.get_question()
            predicate = question.get_predicate().get_name()
            intent = "%s::%s" % (QUESTION_INTENT, predicate)
            result[intent] = list(examples(question))
        return result

    def _examples_of_question(self, grammar, question):
        predicate = question.get_predicate().get_name()
        questions = grammar.questions_of_predicate(predicate)
        for question in questions:
            for example in self._examples_of_intent(grammar, question):
                yield example

    def _examples_of_answers(self, grammar):
        def examples():
            for answer in grammar.answers():
                for example in self._examples_of_intent(grammar, answer):
                    yield example

        return {ANSWER_INTENT: list(examples())}

    def _examples_of_sortal_answers_from_individuals(self, grammar):
        def examples():
            for sort in self._ddd.ontology.get_sorts().values():
                examples = self._examples_of_sortal_answers_of_kind(grammar, sort, self._answer_templates)
                for example in examples:
                    yield example

        return {ANSWER_INTENT: list(examples())}

    def _examples_of_sortal_answers_of_kind(self, grammar, sort, intent_templates):
        sort_should_be_generated_as_entity = sort.is_string_sort() or not sort.is_builtin()
        entity_template = SORTAL_ENTITY_TEMPLATE if sort_should_be_generated_as_entity else BUILTIN_ENTITY_TEMPLATE
        for grammar_entries in self._individual_grammar_entries_samples(grammar, sort):
            examples = self._examples_of_individual(grammar_entries, sort.get_name(), intent_templates, entity_template)
            for example in examples:
                yield example

    def _examples_of_sortal_answer_negations_from_individuals(self, grammar):
        def examples():
            for sort in self._ddd.ontology.get_sorts().values():
                if sort.is_string_sort():
                    continue
                examples = self._examples_of_sortal_answers_of_kind(grammar, sort, self._answer_negation_templates)
                for example in examples:
                    yield example

        return {ANSWER_NEGATION_INTENT: list(examples())}

    @property
    def _answer_templates(self):
        template = Template('{{ name }}')
        return [template]

    @property
    def _answer_negation_templates(self):
        template = Template('not {{ name }}')
        return [template]

    def _examples_of_individual(self, grammar_entries, identifier, intent_templates, entity_template):
        for grammar_entry in grammar_entries:
            examples = self._examples_from_templates(grammar_entry, identifier, intent_templates, entity_template)
            for example in examples:
                yield example

    def _examples_from_templates(self, grammar_entry, identifier, intent_templates, entity_template):
        for intent_template in intent_templates:
            entity = entity_template.render(grammar_entry=grammar_entry, value=identifier)
            yield intent_template.render(name=entity)



# -*- coding: utf-8 -*-

import os
import re
import shutil
import tempfile
import unittest

from mock import MagicMock, Mock, patch

from tala.model.ddd import DDD
from tala.model.grammar.grammar import GrammarBase
from tala.model.grammar.intent import Request, Question, Answer
from tala.model.grammar.required_entity import RequiredPropositionalEntity, RequiredSortalEntity
from tala.nl.languages import ENGLISH
from tala.model.domain import Domain
from tala.model.goal import ResolveGoal
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model.ontology import Ontology
from tala.model.predicate import Predicate
from tala.model.question import WhQuestion
from tala.model.sort import Sort, CustomSort
from tala.nl.rasa.generating import generator
from tala.nl.rasa.generating.generator import RasaGenerator
from tala.nl.rasa.generating.examples import Examples, SortNotSupportedException


class GeneratorTestsBase(object):
    def setup(self):
        self._temp_dir = tempfile.mkdtemp(prefix="GeneratorTests")
        self._cwd = os.getcwd()
        os.chdir(self._temp_dir)
        self._mocked_ddd = self._create_mocked_ddd()
        self._generator = None
        self._mocked_grammar = None
        self._grammar_reader_patcher = self._create_grammar_reader_patcher()
        self._result = None
        self._mocked_warnings = None

    def tearDown(self):
        os.chdir(self._cwd)
        shutil.rmtree(self._temp_dir)
        self._grammar_reader_patcher.stop()

    def _create_mocked_ddd(self):
        ontology = Mock(spec=Ontology)
        ontology.get_individuals_of_sort.return_value = list()
        ontology.get_ddd_specific_actions.return_value = set()
        ontology.get_sorts.return_value = dict()
        ontology.get_predicates.return_value = dict()
        ontology.get_predicate.side_effect = Exception()
        domain = Mock(spec=Domain)
        domain.get_all_resolve_goals.return_value = []
        ddd = Mock(spec=DDD)
        ddd.ontology = ontology
        ddd.domain = domain
        ddd.grammars = {}
        return ddd

    def _create_grammar_reader_patcher(self):
        patcher = patch("%s.GrammarReader" % generator.__name__, autospec=True)
        MockGrammarReader = patcher.start()
        MockGrammarReader.xml_grammar_exists_for_language.return_value = True
        return patcher

    def given_ddd_name(self, name):
        self._mocked_ddd.name = name

    def given_ontology(
        self,
        sort,
        individuals,
        predicate,
        is_builtin=False,
        is_integer_sort=False,
        is_string_sort=False,
        is_real_sort=False,
        is_datetime_sort=False
    ):
        mocked_sort = Mock(spec=Sort)
        mocked_sort.get_name.return_value = sort
        mocked_sort.is_builtin.return_value = is_builtin
        mocked_sort.is_integer_sort.return_value = is_integer_sort
        mocked_sort.is_string_sort.return_value = is_string_sort
        mocked_sort.is_real_sort.return_value = is_real_sort
        mocked_sort.is_domain_sort.return_value = False
        mocked_sort.is_datetime_sort.return_value = is_datetime_sort
        self._mocked_ddd.ontology.individual_sort.return_value = mocked_sort
        self._mocked_ddd.ontology.get_individuals_of_sort.return_value = individuals
        self._mocked_ddd.ontology.get_sorts.return_value = {sort: mocked_sort}
        self._mocked_ddd.ontology.get_sort.return_value = mocked_sort
        mocked_predicate = Mock(spec=Predicate)
        mocked_predicate.get_name.return_value = predicate
        mocked_predicate.getSort.return_value = mocked_sort
        self._mocked_ddd.ontology.get_predicates.return_value = {predicate: mocked_predicate}
        self._mocked_ddd.ontology.get_predicate.side_effect = None
        self._mocked_ddd.ontology.get_predicate.return_value = mocked_predicate

    def given_actions_in_ontology(self, actions):
        self._mocked_ddd.ontology.get_ddd_specific_actions.return_value = actions

    def given_mocked_grammar(self, requests=None, questions=None, individuals=None, answers=None, strings=None):
        self._mocked_grammar = Mock(spec=GrammarBase)
        self._mocked_grammar.requests_of_action.return_value = requests or []
        self._mocked_grammar.questions_of_predicate.return_value = questions or []
        self._mocked_grammar.answers.return_value = answers or []
        self._mocked_grammar.strings_of_predicate.return_value = strings or []
        individuals = individuals or {}

        def get_individual(name):
            return individuals[name]

        self._mocked_grammar.entries_of_individual.side_effect = get_individual
        self._mocked_ddd.grammars["eng"] = self._mocked_grammar

    def given_generator(self):
        self._generator = RasaGenerator(self._mocked_ddd, ENGLISH)

    def when_generate(self):
        self._result = self._generate()

    def _generate(self):
        return self._generator.generate()

    def when_generating_data_with_mocked_grammar_then_exception_is_raised_matching(
        self, expected_exception, expected_pattern
    ):
        try:
            self._generate()
            assert False, "%s not raised" % expected_exception
        except expected_exception as e:
            assert re.match(expected_pattern, str(e))

    def given_expected_plan_questions_in_domain(self, predicates):
        def resolve_goals_of_questions(questions):
            for question in questions:
                mocked_goal = Mock(spec=ResolveGoal)
                mocked_goal.get_question.return_value = question
                yield mocked_goal

        plan_questions = list(self._plan_questions(predicates))
        mocked_resolve_goals = list(resolve_goals_of_questions(plan_questions))
        self._mocked_ddd.domain.get_all_resolve_goals.return_value = mocked_resolve_goals

    def _plan_questions(self, predicates):
        for predicate_name, sort_name in predicates.iteritems():
            predicate = Predicate("mocked_ontology", predicate_name, CustomSort("mocked_ontology", sort_name))
            proposition = LambdaAbstractedPredicateProposition(predicate, "mocked_ontology")
            question = WhQuestion(proposition)
            yield question

    def then_result_matches(self, expected_contents):
        print self._result
        expected_pattern = re.escape(expected_contents)
        assert re.search(expected_pattern, self._result, re.UNICODE) is not None

    def given_mocked_warnings(self, mock_warnings):
        self._mocked_warnings = mock_warnings

    def then_warning_is_issued(self, expected_message):
        self._mocked_warnings.warn.assert_called_once_with(expected_message, UserWarning)


class UnsupportedBuiltinSortGeneratorTestCase(GeneratorTestsBase, unittest.TestCase):
    def setUp(self):
        GeneratorTestsBase.setup(self)

    def test_generate_requests(self):
        self.given_ddd_name("mocked_ddd")
        self.given_ontology(sort="real", predicate="selected_price", individuals=[], is_builtin=True, is_real_sort=True)
        self.given_actions_in_ontology({"purchase"})
        self.given_mocked_grammar(requests=self._requests_of_action("purchase", "selected_price"))
        self.given_generator()
        self.when_generating_data_with_mocked_grammar_then_exception_is_raised_matching(
            SortNotSupportedException, "Builtin sort 'real' is not yet supported together with RASA"
        )

    def _requests_of_action(self, action, required_predicate):
        return [
            Request(action, ["take a note that ", ""], [
                RequiredPropositionalEntity(required_predicate),
            ]),
        ]

    def test_generate_questions(self):
        self.given_ddd_name("mocked_ddd")
        self.given_ontology(sort="real", predicate="selected_price", individuals=[], is_builtin=True, is_real_sort=True)
        self.given_expected_plan_questions_in_domain({"selected_price": "real"})
        self.given_mocked_grammar(questions=self._questions_of_predicate("selected_price", "real"))
        self.given_generator()
        self.when_generating_data_with_mocked_grammar_then_exception_is_raised_matching(
            SortNotSupportedException, "Builtin sort 'real' is not yet supported together with RASA"
        )

    def _questions_of_predicate(self, question_predicate, sort):
        return [
            Question(
                question_predicate, ["how long time remains of the ", " reminder"], [
                    RequiredSortalEntity(sort),
                ]
            ),
        ]

    def test_generate_answers(self):
        self.given_ddd_name("mocked_ddd")
        self.given_ontology(sort="real", predicate="selected_price", individuals=[], is_builtin=True, is_real_sort=True)
        self.given_mocked_grammar()
        self.given_generator()
        self.when_generating_data_with_mocked_grammar_then_exception_is_raised_matching(
            SortNotSupportedException, "Builtin sort 'real' is not yet supported together with RASA"
        )


class CustomSortGeneratorTestCase(GeneratorTestsBase, unittest.TestCase):
    def setUp(self):
        GeneratorTestsBase.setup(self)

    def given_ontology_with_individuals(
        self, sort, predicate, is_integer_sort=False, is_string_sort=False, is_real_sort=False
    ):
        individuals = [
            "contact_john",
            "contact_john_chi",
            "contact_lisa",
            "contact_mary",
            "contact_andy",
            "contact_andy_chi",
        ]
        super(CustomSortGeneratorTestCase,
              self).given_ontology(sort, individuals, predicate, is_integer_sort, is_string_sort, is_real_sort)

    def given_mocked_grammar_with_individuals(self, requests=None, questions=None, answers=None):
        individuals = {
            "contact_john": ["John", "Johnny"],
            "contact_john_chi": [u"约翰"],
            "contact_lisa": ["Lisa", "Elizabeth"],
            "contact_mary": ["Mary"],
            "contact_andy": ["Andy"],
            "contact_andy_chi": [u"安迪"],
        }
        super(CustomSortGeneratorTestCase, self).given_mocked_grammar(requests, questions, individuals, answers)

    def test_generate_requests(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_actions_in_ontology({"call"})
        self.given_mocked_grammar_with_individuals(requests=self._requests_of_action(sort="call"))
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:action::call
- make a call
- call [John](sort.contact)
- call [Johnny](sort.contact)
- call [约翰](sort.contact)
- call [Lisa](sort.contact)
- call [Elizabeth](sort.contact)
- call [Mary](sort.contact)
- call [Andy](sort.contact)
- call [安迪](sort.contact)
"""
        )

    def _requests_of_action(self, sort=None, predicate=None):
        yield Request("call", ["make a call"], [])
        if sort is not None:
            yield Request("call", ["call ", ""], [
                RequiredSortalEntity("contact"),
            ])
        if predicate is not None:
            yield Request("call", ["make a call to ", ""], [
                RequiredPropositionalEntity("selected_contact_to_call"),
            ])

    def test_propositional_entities_excluded_from_requests(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_actions_in_ontology({"call"})
        self.given_mocked_grammar_with_individuals(
            requests=self._requests_of_action(predicate="selected_contact_to_call")
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(u"""## intent:rasa_test:action::call
- make a call

""")

    @patch("{}.warnings".format(generator.__name__), autospec=True)
    def test_requests_with_propositional_entities_issue_warning(self, mock_warnings):
        self.given_mocked_warnings(mock_warnings)
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_actions_in_ontology({"call"})
        self.given_mocked_grammar_with_individuals(
            requests=self._requests_of_action(predicate="selected_contact_to_call")
        )
        self.given_generator()
        self.when_generate()
        self.then_warning_is_issued(
            "Expected only sortal slots but got a propositional slot for predicate "
            "'selected_contact_to_call'. Skipping this training data example."
        )

    def test_generate_requests_with_two_answers(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="caller")
        self.given_actions_in_ontology({"call"})
        self.given_mocked_grammar_with_individuals(
            requests=[
                Request(
                    "call", ["call ", " and say hi from ", ""], [
                        RequiredSortalEntity("contact"),
                        RequiredSortalEntity("contact"),
                    ]
                ),
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:action::call
- call [John](sort.contact) and say hi from [John](sort.contact)
- call [John](sort.contact) and say hi from [Johnny](sort.contact)
- call [John](sort.contact) and say hi from [约翰](sort.contact)
- call [John](sort.contact) and say hi from [Lisa](sort.contact)
- call [John](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [John](sort.contact) and say hi from [Mary](sort.contact)
- call [John](sort.contact) and say hi from [Andy](sort.contact)
- call [John](sort.contact) and say hi from [安迪](sort.contact)
- call [Johnny](sort.contact) and say hi from [John](sort.contact)
- call [Johnny](sort.contact) and say hi from [Johnny](sort.contact)
- call [Johnny](sort.contact) and say hi from [约翰](sort.contact)
- call [Johnny](sort.contact) and say hi from [Lisa](sort.contact)
- call [Johnny](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [Johnny](sort.contact) and say hi from [Mary](sort.contact)
- call [Johnny](sort.contact) and say hi from [Andy](sort.contact)
- call [Johnny](sort.contact) and say hi from [安迪](sort.contact)
- call [约翰](sort.contact) and say hi from [John](sort.contact)
- call [约翰](sort.contact) and say hi from [Johnny](sort.contact)
- call [约翰](sort.contact) and say hi from [约翰](sort.contact)
- call [约翰](sort.contact) and say hi from [Lisa](sort.contact)
- call [约翰](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [约翰](sort.contact) and say hi from [Mary](sort.contact)
- call [约翰](sort.contact) and say hi from [Andy](sort.contact)
- call [约翰](sort.contact) and say hi from [安迪](sort.contact)
- call [Lisa](sort.contact) and say hi from [John](sort.contact)
- call [Lisa](sort.contact) and say hi from [Johnny](sort.contact)
- call [Lisa](sort.contact) and say hi from [约翰](sort.contact)
- call [Lisa](sort.contact) and say hi from [Lisa](sort.contact)
- call [Lisa](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [Lisa](sort.contact) and say hi from [Mary](sort.contact)
- call [Lisa](sort.contact) and say hi from [Andy](sort.contact)
- call [Lisa](sort.contact) and say hi from [安迪](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [John](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [Johnny](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [约翰](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [Lisa](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [Mary](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [Andy](sort.contact)
- call [Elizabeth](sort.contact) and say hi from [安迪](sort.contact)
- call [Mary](sort.contact) and say hi from [John](sort.contact)
- call [Mary](sort.contact) and say hi from [Johnny](sort.contact)
- call [Mary](sort.contact) and say hi from [约翰](sort.contact)
- call [Mary](sort.contact) and say hi from [Lisa](sort.contact)
- call [Mary](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [Mary](sort.contact) and say hi from [Mary](sort.contact)
- call [Mary](sort.contact) and say hi from [Andy](sort.contact)
- call [Mary](sort.contact) and say hi from [安迪](sort.contact)
- call [Andy](sort.contact) and say hi from [John](sort.contact)
- call [Andy](sort.contact) and say hi from [Johnny](sort.contact)
- call [Andy](sort.contact) and say hi from [约翰](sort.contact)
- call [Andy](sort.contact) and say hi from [Lisa](sort.contact)
- call [Andy](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [Andy](sort.contact) and say hi from [Mary](sort.contact)
- call [Andy](sort.contact) and say hi from [Andy](sort.contact)
- call [Andy](sort.contact) and say hi from [安迪](sort.contact)
- call [安迪](sort.contact) and say hi from [John](sort.contact)
- call [安迪](sort.contact) and say hi from [Johnny](sort.contact)
- call [安迪](sort.contact) and say hi from [约翰](sort.contact)
- call [安迪](sort.contact) and say hi from [Lisa](sort.contact)
- call [安迪](sort.contact) and say hi from [Elizabeth](sort.contact)
- call [安迪](sort.contact) and say hi from [Mary](sort.contact)
- call [安迪](sort.contact) and say hi from [Andy](sort.contact)
- call [安迪](sort.contact) and say hi from [安迪](sort.contact)
"""
        )

    def test_generate_questions(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_expected_plan_questions_in_domain({"phone_number_of_contact": "contact"})
        self.given_mocked_grammar_with_individuals(questions=self._questions_of_predicate(sort="contact"))
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:question::phone_number_of_contact
- tell me a phone number
- what is [John](sort.contact)'s number
- what is [Johnny](sort.contact)'s number
- what is [约翰](sort.contact)'s number
- what is [Lisa](sort.contact)'s number
- what is [Elizabeth](sort.contact)'s number
- what is [Mary](sort.contact)'s number
- what is [Andy](sort.contact)'s number
- what is [安迪](sort.contact)'s number
"""
        )

    def _questions_of_predicate(self, sort=None, predicate=None):
        yield Question("phone_number_of_contact", ["tell me a phone number"], [])
        if sort is not None:
            yield Question("phone_number_of_contact", ["what is ", "'s number"], [
                RequiredSortalEntity(sort),
            ])
        if predicate is not None:
            yield Question(
                "phone_number_of_contact", ["tell me ", "'s number"], [
                    RequiredPropositionalEntity(predicate),
                ]
            )

    def test_propositional_entities_excluded_from_questions(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_expected_plan_questions_in_domain({"phone_number_of_contact": "contact"})
        self.given_mocked_grammar_with_individuals(
            questions=self._questions_of_predicate(predicate="selected_contact_to_call")
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:question::phone_number_of_contact
- tell me a phone number

"""
        )

    @patch("{}.warnings".format(generator.__name__), autospec=True)
    def test_questions_with_propositional_entities_issue_warning(self, mock_warnings):
        self.given_mocked_warnings(mock_warnings)
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_expected_plan_questions_in_domain({"phone_number_of_contact": "contact"})
        self.given_mocked_grammar_with_individuals(
            questions=self._questions_of_predicate(predicate="selected_contact_to_call")
        )
        self.given_generator()
        self.when_generate()
        self.then_warning_is_issued(
            "Expected only sortal slots but got a propositional slot for predicate "
            "'selected_contact_to_call'. Skipping this training data example."
        )

    def test_generate_answer_intents(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_mocked_grammar_with_individuals()
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer
- [John](sort.contact)
- [Johnny](sort.contact)
- [约翰](sort.contact)
- [Lisa](sort.contact)
- [Elizabeth](sort.contact)
- [Mary](sort.contact)
- [Andy](sort.contact)
- [安迪](sort.contact)
"""
        )

    def test_generate_answer_negation_intents(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact_to_call")
        self.given_mocked_grammar_with_individuals()
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer_negation
- not [John](sort.contact)
- not [Johnny](sort.contact)
- not [约翰](sort.contact)
- not [Lisa](sort.contact)
- not [Elizabeth](sort.contact)
- not [Mary](sort.contact)
- not [Andy](sort.contact)
- not [安迪](sort.contact)
"""
        )

    @property
    def _contact_data(self):
        return ["Andy", "Mary", "Lisa", "Elizabeth", u"安迪", u"约翰", "John", "Johnny"]

    def test_answers(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact")
        self.given_mocked_grammar_with_individuals(answers=list(self._answers(sort="contact")))
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer
- [John](sort.contact)
- [Johnny](sort.contact)
- [约翰](sort.contact)
- [Lisa](sort.contact)
- [Elizabeth](sort.contact)
- [Mary](sort.contact)
- [Andy](sort.contact)
- [安迪](sort.contact)
- my sortal friend [John](sort.contact)
- my sortal friend [Johnny](sort.contact)
- my sortal friend [约翰](sort.contact)
- my sortal friend [Lisa](sort.contact)
- my sortal friend [Elizabeth](sort.contact)
- my sortal friend [Mary](sort.contact)
- my sortal friend [Andy](sort.contact)
- my sortal friend [安迪](sort.contact)
"""
        )

    def test_propositional_entities_excluded_from_answers(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact")
        self.given_mocked_grammar_with_individuals(answers=list(self._answers(predicate="selected_contact")))
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer
- [John](sort.contact)
- [Johnny](sort.contact)
- [约翰](sort.contact)
- [Lisa](sort.contact)
- [Elizabeth](sort.contact)
- [Mary](sort.contact)
- [Andy](sort.contact)
- [安迪](sort.contact)

"""
        )

    def _answers(self, sort=None, predicate=None):
        if sort is not None:
            yield Answer(["my sortal friend ", ""], [
                RequiredSortalEntity(sort),
            ])
        if predicate is not None:
            yield Answer(["my friend ", ""], [
                RequiredPropositionalEntity(predicate),
            ])

    @patch("{}.warnings".format(generator.__name__), autospec=True)
    def test_answers_with_propositional_entities_issues_warning(self, mock_warnings):
        self.given_mocked_warnings(mock_warnings)
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact")
        self.given_mocked_grammar_with_individuals(answers=list(self._answers(predicate="selected_contact")))
        self.given_generator()
        self.when_generate()
        self.then_warning_is_issued(
            "Expected only sortal slots but got a propositional slot for predicate 'selected_contact'."
            " Skipping this training data example."
        )

    def test_synonyms(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology_with_individuals(sort="contact", predicate="selected_contact")
        self.given_mocked_grammar_with_individuals(answers=list(self._answers(sort="contact")))
        self.given_generator()
        self.when_generate()
        self.then_result_matches(u"""## synonyms:rasa_test:John
- Johnny

## synonyms:rasa_test:Lisa
- Elizabeth
""")


class BuiltinSortGeneratorTestCase(GeneratorTestsBase, unittest.TestCase):
    def setUp(self):
        GeneratorTestsBase.setup(self)
        self._mock_builtin_sort_examples = ["mock example 1", "mock example 2"]
        self._examples_patcher = self._create_examples_patcher()

    def tearDown(self):
        GeneratorTestsBase.tearDown(self)
        self._examples_patcher.stop()

    def _create_examples_patcher(self):
        def create_mock_examples():
            mock_examples = MagicMock(spec=Examples)
            mock_examples.get_builtin_sort_examples.return_value = self._mock_builtin_sort_examples
            return mock_examples

        patcher = patch("%s.Examples" % generator.__name__)
        mock_examples_patcher = patcher.start()
        mock_examples_patcher.from_language.return_value = create_mock_examples()
        return patcher

    def test_generate_requests(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_actions_in_ontology({"mock_action"})
        self.given_mocked_grammar(
            requests=[
                Request("mock_action", ["mock phrase without entities"], []),
                Request("mock_action", ["mock phrase with sortal entity ", ""], [RequiredSortalEntity("mock_sort")]),
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:action::mock_action
- mock phrase without entities
- mock phrase with sortal entity mock example 1
- mock phrase with sortal entity mock example 2
"""
        )

    def test_propositional_entities_excluded_from_requests(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_actions_in_ontology({"mock_action"})
        self.given_mocked_grammar(
            requests=[
                Request("mock_action", ["mock phrase without entities"], []),
                Request(
                    "mock_action", ["mock phrase with propositional entity ", ""],
                    [RequiredPropositionalEntity("mock_predicate")]
                )
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(u"""## intent:rasa_test:action::mock_action
- mock phrase without entities

""")

    @patch("{}.warnings".format(generator.__name__), autospec=True)
    def test_requests_with_propositional_entities_issues_warning(self, mock_warnings):
        self.given_mocked_warnings(mock_warnings)
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_actions_in_ontology({"mock_action"})
        self.given_mocked_grammar(
            requests=[
                Request("mock_action", ["mock phrase without entities"], []),
                Request(
                    "mock_action", ["mock phrase with propositional entity ", ""],
                    [RequiredPropositionalEntity("mock_predicate")]
                )
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_warning_is_issued(
            "Expected only sortal slots but got a propositional slot for predicate 'mock_predicate'."
            " Skipping this training data example."
        )

    def test_generate_questions(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_expected_plan_questions_in_domain(predicates={"mock_predicate": "mock_sort"})
        self.given_mocked_grammar(
            questions=[
                Question("mock_predicate", ["mock phrase without entities"], []),
                Question(
                    "mock_predicate", ["mock phrase with sortal entity ", ""], [RequiredSortalEntity("mock_sort")]
                ),
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:question::mock_predicate
- mock phrase without entities
- mock phrase with sortal entity mock example 1
- mock phrase with sortal entity mock example 2
"""
        )

    def test_propositional_entities_excluded_from_questions(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_expected_plan_questions_in_domain(predicates={"mock_predicate": "mock_sort"})
        self.given_mocked_grammar(
            questions=[
                Question("mock_predicate", ["mock phrase without entities"], []),
                Question(
                    "mock_predicate", ["mock phrase with propositional entity ", ""],
                    [RequiredPropositionalEntity("mock_predicate")]
                )
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(u"""## intent:rasa_test:question::mock_predicate
- mock phrase without entities

""")

    @patch("{}.warnings".format(generator.__name__), autospec=True)
    def test_questions_with_propositional_entities_issues_warning(self, mock_warnings):
        self.given_mocked_warnings(mock_warnings)
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_expected_plan_questions_in_domain(predicates={"mock_predicate": "mock_sort"})
        self.given_mocked_grammar(
            questions=[
                Question("mock_predicate", ["mock phrase without entities"], []),
                Question(
                    "mock_predicate", ["mock phrase with propositional entity ", ""],
                    [RequiredPropositionalEntity("mock_predicate")]
                )
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_warning_is_issued(
            "Expected only sortal slots but got a propositional slot for predicate 'mock_predicate'."
            " Skipping this training data example."
        )

    def test_generate_answer_intents(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_mocked_grammar()
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer
- mock example 1
- mock example 2
"""
        )

    def test_generate_answer_negation_intents(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_mocked_grammar()
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer_negation
- not mock example 1
- not mock example 2
"""
        )

    def test_propositional_entities_excluded_from_answers(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_mocked_grammar(
            answers=[
                Answer(["mock phrase with propositional entity ", ""], [RequiredPropositionalEntity("mock_predicate")])
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:answer
- mock example 1
- mock example 2

"""
        )

    @patch("{}.warnings".format(generator.__name__), autospec=True)
    def test_answers_with_propositional_entities_issues_warning(self, mock_warnings):
        self.given_mocked_warnings(mock_warnings)
        self.given_ddd_name("rasa_test")
        self.given_ontology(sort="mock_sort", individuals=[], predicate="mock_predicate", is_builtin=True)
        self.given_mocked_grammar(
            answers=[
                Answer(["mock phrase with propositional entity ", ""], [RequiredPropositionalEntity("mock_predicate")])
            ]
        )
        self.given_generator()
        self.when_generate()
        self.then_warning_is_issued(
            "Expected only sortal slots but got a propositional slot for predicate 'mock_predicate'."
            " Skipping this training data example."
        )


class StringSortGeneratorTestCase(GeneratorTestsBase, unittest.TestCase):
    def setUp(self):
        GeneratorTestsBase.setup(self)

    def test_examples_extended_with_strings_of_predicate(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(
            sort="string", predicate="mock_predicate", individuals=[], is_builtin=True, is_string_sort=True
        )
        self.given_actions_in_ontology({"mock_action"})
        self.given_mocked_grammar(
            requests=[
                Request("mock_action", ["mock phrase without entities"], []),
                Request("mock_action", ["mock phrase with sortal entity ", ""], [RequiredSortalEntity("string")]),
                Request(
                    "mock_action", ["mock phrase with propositional entity ", ""],
                    [RequiredPropositionalEntity("mock_predicate")]
                )
            ],
            strings=["mock string of predicate"]
        )
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            u"""## intent:rasa_test:action::mock_action
- mock phrase without entities
- mock phrase with sortal entity [single](sort.string)
- mock phrase with sortal entity [double word](sort.string)
- mock phrase with sortal entity [three in one](sort.string)
- mock phrase with sortal entity [hey make it four](sort.string)
- mock phrase with sortal entity [the more the merrier five](sort.string)
- mock phrase with sortal entity [calm down and count to six](sort.string)
- mock phrase with sortal entity [bring them through to the jolly seven](sort.string)
- mock phrase with sortal entity [noone counts toes like an eight toed guy](sort.string)
- mock phrase with sortal entity [it matters to make sense for nine of us](sort.string)
- mock phrase with sortal entity [would you bring ten or none to a desert island](sort.string)
- mock phrase with propositional entity [single](predicate.mock_predicate)
- mock phrase with propositional entity [double word](predicate.mock_predicate)
- mock phrase with propositional entity [three in one](predicate.mock_predicate)
- mock phrase with propositional entity [hey make it four](predicate.mock_predicate)
- mock phrase with propositional entity [the more the merrier five](predicate.mock_predicate)
- mock phrase with propositional entity [calm down and count to six](predicate.mock_predicate)
- mock phrase with propositional entity [bring them through to the jolly seven](predicate.mock_predicate)
- mock phrase with propositional entity [noone counts toes like an eight toed guy](predicate.mock_predicate)
- mock phrase with propositional entity [it matters to make sense for nine of us](predicate.mock_predicate)
- mock phrase with propositional entity [would you bring ten or none to a desert island](predicate.mock_predicate)
- mock phrase with propositional entity [mock string of predicate](predicate.mock_predicate)
"""
        )

    def test_do_not_generate_answer_negation_intents(self):
        self.given_ddd_name("rasa_test")
        self.given_ontology(
            sort="string", predicate="selected_message", individuals=[], is_builtin=True, is_string_sort=True
        )
        self.given_mocked_grammar()
        self.given_generator()
        self.when_generate()
        self.then_result_does_not_match(u"intent:rasa_test:answer_negation")

    def then_result_does_not_match(self, contents):
        expected_pattern = re.escape(contents)
        assert re.search(expected_pattern, self._result, re.UNICODE) is None


class NegativeIntentGeneratorTestCase(GeneratorTestsBase, unittest.TestCase):
    def setUp(self):
        GeneratorTestsBase.setup(self)

    def test_negative_intent(self):
        self.given_ddd_name("mocked_ddd")
        self.given_ontology(sort="mocked_sort", predicate="mocked_predicate", individuals=[])
        self.given_mocked_grammar()
        self.given_generator()
        self.when_generate()
        self.then_result_matches(
            """## intent:NEGATIVE
- aboard
- about
- above
"""
        )



# -*- coding: utf-8 -*-

import pytest
import unittest

from tala.nl.gf import utils

NL_PLACEHOLDER = utils.nl_user_answer_placeholder_of_sort("mock_sort", 5)
NL_PLACEHOLDER_2 = utils.nl_user_answer_placeholder_of_sort("mocksort2", 99)

SEM_PLACEHOLDER = utils.semantic_user_answer_placeholder_of_sort("mock_sort", 5)
SEM_PLACEHOLDER_2 = utils.semantic_user_answer_placeholder_of_sort("mocksort2", 99)


@pytest.mark.parametrize(
    "unformatted_string, placeholder, empty_string", [
        ["zldknfb %s", NL_PLACEHOLDER, ""],
        ["zldknfb %s lsdgfnkj", NL_PLACEHOLDER, ""],
        ["zldknfb %s %s lsdgfnkj", (NL_PLACEHOLDER, NL_PLACEHOLDER_2), ("", "")],
    ])
def test_remove_nl_placeholder_from_string(unformatted_string, placeholder, empty_string):
    original = unformatted_string % placeholder
    expected = unformatted_string % empty_string
    actual = utils.remove_nl_placeholders_from_string(original)
    assert expected.split() == actual.split()


@pytest.mark.parametrize(
    "unformatted_string, placeholder, empty_string", [
        ["zldknfb %s", SEM_PLACEHOLDER, ""],
        ["zldknfb %s lsdgfnkj", SEM_PLACEHOLDER, ""],
        ["zldknfb %s %s lsdgfnkj", (SEM_PLACEHOLDER, SEM_PLACEHOLDER_2), ("", "")],
    ])
def test_remove_sem_placeholder_from_string(unformatted_string, placeholder, empty_string):
    original = unformatted_string % placeholder
    expected = unformatted_string % empty_string
    actual = utils.remove_sem_placeholders_from_string(original)
    assert expected.split() == actual.split()


class BindTokensTestCase(unittest.TestCase):
    def test_base_case(self):
        self._when_bind_tokens(["what", utils.BIND, "'s", "the", "time"])
        self._then_result_is(["what's", "the", "time"])

    def test_no_bind(self):
        self._when_bind_tokens(["what", "is", "the", "time"])
        self._then_result_is(["what", "is", "the", "time"])

    def test_multiple_binds(self):
        self._when_bind_tokens(["what", utils.BIND, "'s", "john", utils.BIND, "'s", "number"])
        self._then_result_is(["what's", "john's", "number"])

    def _when_bind_tokens(self, tokens):
        self._result = utils.bind_tokens(tokens)

    def _then_result_is(self, expected_result):
        self.assertEquals(expected_result, self._result)


class StringReplacementTestCase(unittest.TestCase):
    def test_preprocess_strings_double_quotes(self):
        self.when_preprocess_string(
            'answer(comment("a comment"))')
        self.then_expect(
            preprocessed_string="answer(comment(_STR0_))",
            table={"_STR0_": "a comment"})

    def when_preprocess_string(self, string):
        self._string_to_preprocess = string
        self._actual_table = {}
        self._actual_preprocessed_string = utils.preprocess_strings(string, self._actual_table)

    def then_expect(self, preprocessed_string, table):
        self.assertEquals(preprocessed_string, self._actual_preprocessed_string)
        self.assertEquals(table, self._actual_table)

    def test_preprocess_strings_single_quotes(self):
        self.when_preprocess_string(
            "answer(comment('a comment'))")
        self.then_expect(
            preprocessed_string="answer(comment(_STR0_))",
            table={"_STR0_": "a comment"})

    def test_preprocess_strings_mixed_quotes_no_action(self):
        self.when_preprocess_string(
            "answer(comment('10:05\"))")
        self.then_expect_unchanged_string()

    def then_expect_unchanged_string(self):
        self.assertEquals(self._string_to_preprocess, self._actual_preprocessed_string)

    def test_preprocess_multiple_strings(self):
        self.when_preprocess_string(
            'report(ServiceResultProposition(RegisterComment, [comment_message("mock comment"), comment_name("mock name")], SuccessfulServiceAction()))')
        self.then_expect(
            preprocessed_string='report(ServiceResultProposition(RegisterComment, [comment_message(_STR0_), comment_name(_STR1_)], SuccessfulServiceAction()))',
            table={"_STR0_": "mock comment",
                   "_STR1_": "mock name"})


class DateTimeReplacementTestCase(unittest.TestCase):
    def test_preprocess_base_case(self):
        self._when_string_is_preprocessed('report(ServiceResultProposition(BookFlight, [desired_time(datetime(2018-04-11T22:00:00.000Z))], SuccessfulServiceAction()), ddd_name=u\'rasa_time\')')
        self._then_string_and_table_are(
            'report(ServiceResultProposition(BookFlight, [desired_time(_datetime_placeholder_0_)], SuccessfulServiceAction()), ddd_name=u\'rasa_time\')',
            {"_datetime_placeholder_0_": "04/11/2018 10:00 PM"})

    def _when_string_is_preprocessed(self, string):
        self._actual_table = {}
        self._result = utils.preprocess_datetimes(string, self._actual_table)

    def _then_string_and_table_are(self, expected_string, expected_table):
        self.assertEquals(expected_string, self._result)
        self.assertEquals(expected_table, self._actual_table)


class IntegerReplacementTestCase(unittest.TestCase):
    def test_preprocess_integers_doesnt_touch_plain_digits(self):
        self._when_string_is_preprocessed("222")
        self._then_string_and_table_are("222", {})

    def test_preprocess_integers_processes_more_than_2_digits_in_parens(self):
        self._when_string_is_preprocessed("(222)")
        self._then_string_and_table_are("(_integer_placeholder_0_)", {"_integer_placeholder_0_": "222"})

    def test_preprocess_integers_skips_2_or_less_digits_in_parens(self):
        self._when_string_is_preprocessed("(22)")
        self._then_string_and_table_are("(22)", {})

    def _when_string_is_preprocessed(self, string):
        self._actual_table = {}
        self._result = utils.preprocess_integers(string, self._actual_table)

    def _then_string_and_table_are(self, expected_string, expected_table):
        self.assertEquals(expected_string, self._result)
        self.assertEquals(expected_table, self._actual_table)

    def _then_result_is(self, tokens):
        self.assertEquals(tokens, self._result)


class GenericPlaceholderReplacementTestCase(unittest.TestCase):
    def test_base_case(self):
        self._when_replace_placeholders_is_invoked(
            "the comment has _integer_placeholder_0_ likes",
            {"_integer_placeholder_0_": "222"})
        self._then_result_is("the comment has 222 likes")

    def _when_replace_placeholders_is_invoked(self, string, table):
        self._result = utils.replace_placeholders(string, table)

    def _then_result_is(self, expected_result):
        self.assertEquals(expected_result, self._result)

    def test_placeholder_not_as_its_own_token(self):
        self._when_replace_placeholders_is_invoked(
            "_integer_placeholder_0_: this is not a valid temperature",
            {"_integer_placeholder_0_": "222"})
        self._then_result_is("222: this is not a valid temperature")


class str_phrase_test_case(unittest.TestCase):
    def test_base_case(self):
        self.assert_phrase_to_string(
            ["calling", "alex", "berman"],
            "calling alex berman")

    def test_closing_punctuation(self):
        self.assert_phrase_to_string(
            ["okay", "."],
            "okay.")
        self.assert_phrase_to_string(
            ["okay", "?"],
            "okay?")
        self.assert_phrase_to_string(
            ["okay", "!"],
            "okay!")

    def test_comma(self):
        self.assert_phrase_to_string(
            ["okay", ",", "so"],
            "okay, so")

    def test_bind_genitive(self):
        self.assert_phrase_to_string(
            ["berman", utils.BIND, "'", utils.BIND, "s"],
            "berman's")

    def test_bind_between_word_and_comma(self):
        self.assert_phrase_to_string(
            ["berman", utils.BIND, ",", "alex"],
            "berman, alex")

    def assert_phrase_to_string(self, tokens, expected_string):
        self.assertEquals(expected_string, utils.str_phrase(tokens))


class GFTests(unittest.TestCase):
    def test_tokenise(self):
        self.assertEquals(['pred', '(', 'arg', ')'],
                          utils.tokenise('pred(arg)'))

    def test_tokenise_with_string_constant(self):
        self.assertEquals(['you', 'said', '_STR0_', '.'],
                          utils.tokenise('you said _STR0_.'))

    def test_tokenise_genitive(self):
        self.assertEquals(["Alex's"],
                          utils.tokenise("Alex's"))

    def test_tokenise_contraction(self):
        self.assertEquals(["don't"],
                          utils.tokenise("don't"))

    def test_tokenise_time_expressions(self):
        self.assertEquals(["1:30", "10:30"],
                          utils.tokenise("1:30 10:30"))

    def test_tokenise_time_expressions_negative(self):
        self.assertEquals(
            [":", "30", "10:30", "3", "101", ":", "30", "19", ":", "1"],
            utils.tokenise(":30 10:303 101:30 19:1"))

    def test_tokenise_with_magic_wording(self):
        self.assertEquals(["a", "message"],
                          utils.tokenise("a message"))

    def test_tokenise_am_pm(self):
        self.assertEquals(["p.m.", "2:10", "a.m."],
                          utils.tokenise("p.m. 2:10 a.m."))

    def test_tokenise_am_pm_and_magic_wordings(self):
        self.assertEquals(["send", "a", "message", "at", "2", "p.m."],
                          utils.tokenise("send a message at 2 p.m."))

    def test_tokenisation_doesnt_split_at_hyphen(self):
        self.assertEquals(["Berlin-Tegel"], utils.tokenise("Berlin-Tegel"))

    def test_dont_split_7bit_encoded_tokens(self):
        string = u"ställa klockan"
        expected_tokens = [u"ställa", "klockan"]
        self.assertEquals(expected_tokens, utils.tokenise(string))
